@@ -14,12 +14,15 @@ package com.sun.jna.platform.win32;
 
 import java.io.File;
 import java.io.FileNotFoundException;
-import java.util.ArrayList;
+import java.nio.ByteOrder;
 import java.util.List;
+import java.util.Map;
+import java.util.TreeMap;
 
 import com.sun.jna.LastErrorException;
 import com.sun.jna.Memory;
 import com.sun.jna.Native;
+import com.sun.jna.Pointer;
 import com.sun.jna.platform.win32.WinNT.HANDLE;
 import com.sun.jna.platform.win32.WinNT.HANDLEByReference;
 import com.sun.jna.platform.win32.WinNT.HRESULT;
@@ -31,6 +34,7 @@ import com.sun.jna.ptr.PointerByReference;
  * 
  * @author dblock[at]dblock.org
  * @author markus[at]headcrashing[dot]eu
+ * @author Andreas "PAX" L&uuml;ck, onkelpax-git[at]yahoo.de
  */
 public abstract class Kernel32Util implements WinDef {
 
@@ -83,7 +87,11 @@ public abstract class Kernel32Util implements WinDef {
         return formatMessage(code.intValue());
     }
 
-    /** @deprecated use {@link #formatMessage(WinNT.HRESULT)} instead. */
+    /**
+     * @deprecated use {@link #formatMessage(WinNT.HRESULT)} instead.
+     * @param code error code
+     * @return formatted message
+     */
     public static String formatMessageFromHR(HRESULT code) {
         return formatMessage(code.intValue());
     }
@@ -99,6 +107,15 @@ public abstract class Kernel32Util implements WinDef {
         return formatMessageFromHR(W32Errors.HRESULT_FROM_WIN32(code));
     }
 
+	/**
+	 * @return Obtains the human-readable error message text from the last error
+	 *         that occurred by invocating {@code Kernel32.GetLastError()}.
+	 */
+	public static String getLastErrorMessage() {
+		return Kernel32Util.formatMessageFromLastErrorCode(Kernel32.INSTANCE
+				.GetLastError());
+	}
+
     /**
      * Return the path designated for temporary files.
      * 
@@ -122,33 +139,22 @@ public abstract class Kernel32Util implements WinDef {
     /**
      * Returns valid drives in the system.
      * 
-     * @return An array of valid drives.
+     * @return A {@link List} of valid drives.
      */
-    public static String[] getLogicalDriveStrings() {
-        DWORD dwSize = Kernel32.INSTANCE.GetLogicalDriveStrings(new DWORD(0),
-                null);
+    public static List<String> getLogicalDriveStrings() {
+        DWORD dwSize = Kernel32.INSTANCE.GetLogicalDriveStrings(new DWORD(0), null);
         if (dwSize.intValue() <= 0) {
             throw new Win32Exception(Kernel32.INSTANCE.GetLastError());
         }
 
         char buf[] = new char[dwSize.intValue()];
         dwSize = Kernel32.INSTANCE.GetLogicalDriveStrings(dwSize, buf);
-        if (dwSize.intValue() <= 0) {
+        int bufSize = dwSize.intValue();
+        if (bufSize <= 0) {
             throw new Win32Exception(Kernel32.INSTANCE.GetLastError());
         }
 
-        List<String> drives = new ArrayList<String>();
-        String drive = "";
-        // the buffer is double-null-terminated
-        for (int i = 0; i < buf.length - 1; i++) {
-            if (buf[i] == 0) {
-                drives.add(drive);
-                drive = "";
-            } else {
-                drive += buf[i];
-            }
-        }
-        return drives.toArray(new String[0]);
+        return Native.toStringList(buf, 0, bufSize);
     }
 
     /**
@@ -168,6 +174,9 @@ public abstract class Kernel32Util implements WinDef {
 
     /**
      * Retrieves the result of GetFileType, provided the file exists.
+     * @param fileName file name
+     * @return file type
+     * @throws FileNotFoundException if file not found
      */
     public static int getFileType(String fileName) throws FileNotFoundException {
         File f = new File(fileName);
@@ -211,6 +220,7 @@ public abstract class Kernel32Util implements WinDef {
     }
 
     /**
+     * @param rootName name of root drive
      * @return One of the WinBase.DRIVE_* constants.
      */
     public static int getDriveType(String rootName) {
@@ -242,6 +252,177 @@ public abstract class Kernel32Util implements WinDef {
         return Native.toString(buffer);
     }
 
+    /**
+     * Uses the {@link Kernel32#GetEnvironmentStrings()} to retrieve and
+     * parse the current process environment
+     * @return The current process environment as a {@link Map}.
+     * @throws LastErrorException if failed to get or free the environment
+     * data block
+     * @see #getEnvironmentVariables(Pointer, long)
+     */
+    public static Map<String,String> getEnvironmentVariables() {
+        Pointer lpszEnvironmentBlock=Kernel32.INSTANCE.GetEnvironmentStrings();
+        if (lpszEnvironmentBlock == null) {
+            throw new LastErrorException(Kernel32.INSTANCE.GetLastError());
+        }
+        
+        try {
+            return getEnvironmentVariables(lpszEnvironmentBlock, 0L);
+        } finally {
+            if (!Kernel32.INSTANCE.FreeEnvironmentStrings(lpszEnvironmentBlock)) {
+                throw new LastErrorException(Kernel32.INSTANCE.GetLastError());
+            }
+        }
+    }
+
+    /**
+     * @param lpszEnvironmentBlock The environment block as received from the
+     * <A HREF="https://msdn.microsoft.com/en-us/library/windows/desktop/ms683187(v=vs.85).aspx">GetEnvironmentStrings</A>
+     * function
+     * @param offset Offset within the block to parse the data
+     * @return A {@link Map} of the parsed <code>name=value</code> pairs.
+     * <B>Note:</B> if the environment block is {@code null} then {@code null}
+     * is returned instead of an empty map since we want to distinguish
+     * between the case that the data block is {@code null} and when there are
+     * no environment variables (as unlikely as it may be) 
+     */
+    public static Map<String,String> getEnvironmentVariables(Pointer lpszEnvironmentBlock, long offset) {
+        if (lpszEnvironmentBlock == null) {
+            return null;
+        }
+        
+        Map<String,String>  vars=new TreeMap<String,String>();
+        boolean             asWideChars=isWideCharEnvironmentStringBlock(lpszEnvironmentBlock, offset);
+        long                stepFactor=asWideChars ? 2L : 1L;
+        for (long    curOffset=offset; ; ) {
+            String  nvp=readEnvironmentStringBlockEntry(lpszEnvironmentBlock, curOffset, asWideChars);
+            int     len=nvp.length();
+            if (len == 0) { // found the ending '\0'
+                break;
+            }
+
+            int pos=nvp.indexOf('=');
+            if (pos < 0) {
+                throw new IllegalArgumentException("Missing variable value separator in " + nvp);
+            }
+            
+            String  name=nvp.substring(0, pos), value=nvp.substring(pos + 1);
+            vars.put(name, value);
+
+            curOffset += (len + 1 /* skip the ending '\0' */) * stepFactor;
+        }
+        
+        return vars;
+    }
+
+    /**
+     * @param lpszEnvironmentBlock The environment block as received from the
+     * <A HREF="https://msdn.microsoft.com/en-us/library/windows/desktop/ms683187(v=vs.85).aspx">GetEnvironmentStrings</A>
+     * function
+     * @param offset Offset within the block to look for the entry
+     * @param asWideChars If {@code true} then the block contains {@code wchar_t}
+     * instead of &quot;plain old&quot; {@code char}s
+     * @return A {@link String} containing the <code>name=value</code> pair or
+     * empty if reached end of block
+     * @see #isWideCharEnvironmentStringBlock
+     * @see #findEnvironmentStringBlockEntryEnd
+     */
+    public static String readEnvironmentStringBlockEntry(Pointer lpszEnvironmentBlock, long offset, boolean asWideChars) {
+        long endOffset=findEnvironmentStringBlockEntryEnd(lpszEnvironmentBlock, offset, asWideChars);
+        int  dataLen=(int) (endOffset - offset);
+        if (dataLen == 0) {
+            return "";
+        }
+        
+        int         charsLen=asWideChars ? (dataLen / 2) : dataLen;
+        char[]      chars=new char[charsLen];
+        long        curOffset=offset, stepSize=asWideChars ? 2L : 1L;
+        ByteOrder   byteOrder=ByteOrder.nativeOrder();
+        for (int index=0; index < chars.length; index++, curOffset += stepSize) {
+            byte b=lpszEnvironmentBlock.getByte(curOffset);
+            if (asWideChars) {
+                byte x=lpszEnvironmentBlock.getByte(curOffset + 1L);
+                if (ByteOrder.LITTLE_ENDIAN.equals(byteOrder)) {
+                    chars[index] = (char) (((x << Byte.SIZE) & 0xFF00) | (b & 0x00FF));
+                } else {    // unlikely, but handle it
+                    chars[index] = (char) (((b << Byte.SIZE) & 0xFF00) | (x & 0x00FF));
+                }
+            } else {
+                chars[index] = (char) (b & 0x00FF);
+            }
+        }
+        
+        return new String(chars);
+    }
+
+    /**
+     * @param lpszEnvironmentBlock The environment block as received from the
+     * <A HREF="https://msdn.microsoft.com/en-us/library/windows/desktop/ms683187(v=vs.85).aspx">GetEnvironmentStrings</A>
+     * function
+     * @param offset Offset within the block to look for the entry
+     * @param asWideChars If {@code true} then the block contains {@code wchar_t}
+     * instead of &quot;plain old&quot; {@code char}s
+     * @return The offset of the <U>first</U> {@code '\0'} in the data block
+     * starting at the specified offset - can be the start offset itself if empty
+     * string.
+     * @see #isWideCharEnvironmentStringBlock
+     */
+    public static long findEnvironmentStringBlockEntryEnd(Pointer lpszEnvironmentBlock, long offset, boolean asWideChars) {
+        for (long curOffset=offset, stepSize=asWideChars ? 2L : 1L; ; curOffset += stepSize) {
+            byte b=lpszEnvironmentBlock.getByte(curOffset);
+            if (b == 0) {
+                return curOffset;
+            }
+        }
+    }
+
+    /**
+     * <P>Attempts to determine whether the data block uses {@code wchar_t}
+     * instead of &quot;plain old&quot; {@code char}s. It does that by reading
+     * 2 bytes from the specified offset - the character value and its charset
+     * indicator - and examining them as follows:</P>
+     * <UL>
+     *      <LI>
+     *      If the charset indicator is non-zero then it is assumed to be
+     *      a &quot;plain old&quot; {@code char}s data block. <B>Note:</B>
+     *      the assumption is that the environment variable <U>name</U> (at
+     *      least) is ASCII.
+     *      </LI>
+     *      
+     *      <LI>
+     *      Otherwise (i.e., zero charset indicator), it is assumed to be
+     *      a {@code wchar_t} 
+     *      </LI>
+     * </UL>
+     * <B>Note:</B> the code takes into account the {@link ByteOrder} even though
+     * only {@link ByteOrder#LITTLE_ENDIAN} is the likely one
+     * @param lpszEnvironmentBlock The environment block as received from the
+     * <A HREF="https://msdn.microsoft.com/en-us/library/windows/desktop/ms683187(v=vs.85).aspx">GetEnvironmentStrings</A>
+     * function
+     * @param offset offset
+     * @return {@code true} if the block contains {@code wchar_t} instead of
+     * &quot;plain old&quot; {@code char}s
+     */
+    public static boolean isWideCharEnvironmentStringBlock(Pointer lpszEnvironmentBlock, long offset) {
+        byte        b0=lpszEnvironmentBlock.getByte(offset);
+        byte        b1=lpszEnvironmentBlock.getByte(offset + 1L);
+        ByteOrder   byteOrder=ByteOrder.nativeOrder();
+        if (ByteOrder.LITTLE_ENDIAN.equals(byteOrder)) {
+            return isWideCharEnvironmentStringBlock(b1);
+        } else {
+            return isWideCharEnvironmentStringBlock(b0);
+        }
+    }
+    
+    private static boolean isWideCharEnvironmentStringBlock(byte charsetIndicator) {
+        // assume wchar_t for environment variables represents ASCII letters
+        if (charsetIndicator != 0) {
+            return false;
+        } else {
+            return true;
+        }
+    }
+
     /**
      * Retrieves an integer associated with a key in the specified section of an
      * initialization file.
@@ -425,4 +606,76 @@ public abstract class Kernel32Util implements WinDef {
             throw new Win32Exception(Kernel32.INSTANCE.GetLastError());
         }
     }
+
+    /**
+     * Invokes the {@link Kernel32#QueryDosDevice(String, char[], int)} method
+     * and parses the result
+     * @param lpszDeviceName The device name
+     * @param maxTargetSize The work buffer size to use for the query
+     * @return The parsed result
+     */
+    public static final List<String> queryDosDevice(String lpszDeviceName, int maxTargetSize) {
+        char[] lpTargetPath = new char[maxTargetSize];
+        int dwSize = Kernel32.INSTANCE.QueryDosDevice(lpszDeviceName, lpTargetPath, lpTargetPath.length);
+        if (dwSize == 0) {
+            throw new Win32Exception(Kernel32.INSTANCE.GetLastError());
+        }
+
+        return Native.toStringList(lpTargetPath, 0, dwSize);
+    }
+
+    /**
+     * Invokes and parses the result of {@link Kernel32#GetVolumePathNamesForVolumeName(String, char[], int, IntByReference)}
+     * @param lpszVolumeName The volume name
+     * @return The parsed result
+     * @throws Win32Exception If failed to retrieve the required information
+     */
+    public static final List<String> getVolumePathNamesForVolumeName(String lpszVolumeName) {
+        char[] lpszVolumePathNames = new char[WinDef.MAX_PATH + 1];
+        IntByReference lpcchReturnLength = new IntByReference();
+
+        if (!Kernel32.INSTANCE.GetVolumePathNamesForVolumeName(lpszVolumeName, lpszVolumePathNames, lpszVolumePathNames.length, lpcchReturnLength)) {
+            int hr = Kernel32.INSTANCE.GetLastError();
+            if (hr != WinError.ERROR_MORE_DATA) {
+                throw new Win32Exception(hr);
+            }
+            
+            int required = lpcchReturnLength.getValue();
+            lpszVolumePathNames = new char[required];
+            // this time we MUST succeed
+            if (!Kernel32.INSTANCE.GetVolumePathNamesForVolumeName(lpszVolumeName, lpszVolumePathNames, lpszVolumePathNames.length, lpcchReturnLength)) {
+                throw new Win32Exception(Kernel32.INSTANCE.GetLastError());
+            }
+        }
+        
+        int bufSize = lpcchReturnLength.getValue();
+        return Native.toStringList(lpszVolumePathNames, 0, bufSize);
+    }
+
+    // prefix and suffix of a volume GUID path
+    public static final String VOLUME_GUID_PATH_PREFIX = "\\\\?\\Volume{";
+    public static final String VOLUME_GUID_PATH_SUFFIX = "}\\";
+
+    /**
+     * Parses and returns the pure GUID value of a volume name obtained
+     * from {@link Kernel32#FindFirstVolume(char[], int)} or
+     * {@link Kernel32#FindNextVolume} calls
+     * 
+     * @param volumeGUIDPath
+     *              The volume GUID path as returned by one of the above mentioned calls
+     * @return The pure GUID value after stripping the &quot;\\?\&quot; prefix and
+     * removing the trailing backslash.
+     * @throws IllegalArgumentException if bad format encountered
+     * @see <A HREF="https://msdn.microsoft.com/en-us/library/windows/desktop/aa365248(v=vs.85).aspx">Naming a Volume</A>
+     */
+    public static final String extractVolumeGUID(String volumeGUIDPath) {
+        if ((volumeGUIDPath == null)
+         || (volumeGUIDPath.length() <= (VOLUME_GUID_PATH_PREFIX.length() + VOLUME_GUID_PATH_SUFFIX.length()))
+         || (!volumeGUIDPath.startsWith(VOLUME_GUID_PATH_PREFIX))
+         || (!volumeGUIDPath.endsWith(VOLUME_GUID_PATH_SUFFIX))) {
+            throw new IllegalArgumentException("Bad volume GUID path format: " + volumeGUIDPath);
+        }
+        
+        return volumeGUIDPath.substring(VOLUME_GUID_PATH_PREFIX.length(), volumeGUIDPath.length() - VOLUME_GUID_PATH_SUFFIX.length());
+    }
 }
