@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2013 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2007-2015 Timothy Wall, All Rights Reserved
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -61,7 +61,7 @@ import com.sun.jna.Structure.FFIType;
  * figures out these properties based on its enclosing interface.<p>
  * <a name=library_loading></a>
  * <h2>Library Loading</h2>
- * When JNA classes are loaded, the native shared library (jnidispatch) is
+ * <p>When JNA classes are loaded, the native shared library (jnidispatch) is
  * loaded as well.  An attempt is made to load it from the any paths defined
  * in <code>jna.boot.library.path</code> (if defined), then the system library
  * path using {@link System#loadLibrary}, unless <code>jna.nosys=true</code>.
@@ -71,17 +71,21 @@ import com.sun.jna.Structure.FFIType;
  * additional security constraints regarding execution or load of files
  * (SELinux, for example), you should  probably install the native library in
  * an accessible location and configure  your system accordingly, rather than
- * relying on JNA to extract the library  from its own jar file.<p/>
- * To avoid the automatic unpacking (in situations where you want to force a
+ * relying on JNA to extract the library  from its own jar file.</p>
+ * <p>To avoid the automatic unpacking (in situations where you want to force a
  * failure if the JNA native library is not properly installed on the system),
  * set the system property <code>jna.nounpack=true</code>.
- * <p/>
- * NOTE: all native functions are provided within this class to ensure that
+ * </p>
+ * <p>While this class and its corresponding native library are loaded, the 
+ * system property <code>jna.loaded</code> will be set.  The property will be
+ * cleared when native support has been unloaded (i.e. the Native class and
+ * its underlying native support has been GC'd).</p>
+ * <p>NOTE: all native functions are provided within this class to ensure that
  * all other JNA-provided classes and objects are GC'd and/or
  * finalized/disposed before this class is disposed and/or removed from
  * memory (most notably Memory and any other class which by default frees its
- * resources in a finalizer).<p/>
- * <a name=native_library_loading</a>
+ * resources in a finalizer).</p>
+ * <a name=native_library_loading></a>
  * <h2>Native Library Loading</h2>
  * Native libraries loaded via {@link #loadLibrary(Class)} may be found in
  * <a href="NativeLibrary.html#library_search_paths">several locations</a>.
@@ -92,8 +96,8 @@ import com.sun.jna.Structure.FFIType;
 public final class Native implements Version {
 
     public static final String DEFAULT_ENCODING = "utf8";
-    static final boolean DEBUG_LOAD = Boolean.getBoolean("jna.debug_load");
-    static final boolean DEBUG_JNA_LOAD = Boolean.getBoolean("jna.debug_load.jna");
+    public static boolean DEBUG_LOAD = Boolean.getBoolean("jna.debug_load");
+    public static boolean DEBUG_JNA_LOAD = Boolean.getBoolean("jna.debug_load.jna");
 
     // Used by tests, do not remove
     static String jnidispatchPath = null;
@@ -101,6 +105,7 @@ public final class Native implements Version {
     private static Map libraries = new WeakHashMap();
     private static final UncaughtExceptionHandler DEFAULT_HANDLER =
         new UncaughtExceptionHandler() {
+            @Override
             public void uncaughtException(Callback c, Throwable e) {
                 System.err.println("JNA: Callback " + c + " threw the following exception:");
                 e.printStackTrace();
@@ -127,6 +132,10 @@ public final class Native implements Version {
     static final int MAX_ALIGNMENT;
     static final int MAX_PADDING;
 
+    public static float parseVersion(String v) {
+        return Float.parseFloat(v.substring(0, v.lastIndexOf(".")));
+    }
+
     static {
         loadNativeDispatchLibrary();
         POINTER_SIZE = sizeof(TYPE_VOIDP);
@@ -140,8 +149,8 @@ public final class Native implements Version {
         if (Boolean.getBoolean("jna.protected")) {
             setProtected(true);
         }
-        String version = getNativeVersion();
-        if (!VERSION_NATIVE.equals(version)) {
+        float version = parseVersion(getNativeVersion());
+        if (version != parseVersion(VERSION_NATIVE)) {
             String LS = System.getProperty("line.separator");
             throw new Error(LS + LS
                             + "There is an incompatible JNA native library installed on this system" + LS
@@ -154,25 +163,34 @@ public final class Native implements Version {
                             + " - set jna.boot.library.path to include the path to the version of the " + LS
                             + "   jnidispatch library included with the JNA jar file you are using" + LS);
         }
-	MAX_ALIGNMENT = Platform.isSPARC() || Platform.isWindows()
+        MAX_ALIGNMENT = Platform.isSPARC() || Platform.isWindows()
             || (Platform.isLinux() && (Platform.isARM() || Platform.isPPC()))
             || Platform.isAIX()
             || Platform.isAndroid()
             ? 8 : LONG_SIZE;
         MAX_PADDING = (Platform.isMac() && Platform.isPPC()) ? 8 : MAX_ALIGNMENT;
+        System.setProperty("jna.loaded", "true");
     }
 
-    /** Force a dispose when this class is GC'd. */
+    /** Force a dispose when the Native class is GC'd. */
     private static final Object finalizer = new Object() {
+        @Override
         protected void finalize() {
             dispose();
         }
     };
 
-    /** Properly dispose of JNA functionality. */
+    /** Properly dispose of JNA functionality.
+        Called when this class is finalized and also from JNI when
+        JNA's native shared library is unloaded.
+     */
     private static void dispose() {
+        CallbackReference.disposeAll();
+        Memory.disposeAll();
         NativeLibrary.disposeAll();
+        unregisterAll();
         jnidispatchPath = null;
+        System.setProperty("jna.loaded", "false");
     }
 
     /** Remove any automatically unpacked native library.
@@ -228,21 +246,18 @@ public final class Native implements Version {
     public static synchronized native boolean isProtected();
 
     /** This method is obsolete.  The last error value is always preserved.
-     * <p/>
-     * NOTE: The preferred method of obtaining the last error result is
-     * to declare your mapped method to throw {@link LastErrorException}.
-     *
-     * @deprecated Last error is always preserved
+     * @see #getLastError()
+     * @deprecated Last error is always preserved and available via {@link #getLastError()}
      */
+    @Deprecated
     public static void setPreserveLastError(boolean enable) { }
 
     /** Indicates whether the system last error result is preserved
      * after every invocation.  Always returns <code>true</code><p>
-     * NOTE: The preferred method of obtaining the last error result is
-     * to declare your mapped method to throw {@link LastErrorException}.
-     *
-     * @deprecated Last error is always preserved
+     * @see #getLastError()
+     * @deprecated Last error is always preserved and available via {@link #getLastError()}
      */
+    @Deprecated
     public static boolean getPreserveLastError() { return true; }
 
     /** Utility method to get the native window ID for a Java {@link Window}
@@ -297,51 +312,126 @@ public final class Native implements Version {
 
     private static native long _getDirectBufferPointer(Buffer b);
 
-    /** Obtain a Java String from the given native byte array.  If there is
+    /**
+     * Obtain a Java String from the given native byte array.  If there is
      * no NUL terminator, the String will comprise the entire array.  The
      * encoding is obtained from {@link #getDefaultStringEncoding()}.
+     *
+     * @param buf The buffer containing the encoded bytes
+     * @see #toString(byte[], String)
      */
     public static String toString(byte[] buf) {
         return toString(buf, getDefaultStringEncoding());
     }
 
-    /** Obtain a Java String from the given native byte array, using the given
+    /**
+     * Obtain a Java String from the given native byte array, using the given
      * encoding.  If there is no NUL terminator, the String will comprise the
-     * entire array.  If the <code>encoding</code> parameter is null,
-     * the platform default encoding will be used.
+     * entire array.
+     *
+     * @param buf The buffer containing the encoded bytes
+     * @param encoding The encoding name - if {@code null} then the platform
+     * default encoding will be used
      */
     public static String toString(byte[] buf, String encoding) {
-        String s = null;
+        int len = buf.length;
+        // find out the effective length
+        for (int index = 0; index < len; index++) {
+            if (buf[index] == 0) {
+                len = index;
+                break;
+            }
+        }
+
+        if (len == 0) {
+            return "";
+        }
+
         if (encoding != null) {
             try {
-                s = new String(buf, encoding);
+                return new String(buf, 0, len, encoding);
             }
             catch(UnsupportedEncodingException e) {
                 System.err.println("JNA Warning: Encoding '"
                                    + encoding + "' is unsupported");
             }
         }
-        if (s == null) {
-            System.err.println("JNA Warning: Decoding with fallback " + System.getProperty("file.encoding"));
-            s = new String(buf);
-        }
-        int term = s.indexOf(0);
-        if (term != -1) {
-            s = s.substring(0, term);
-        }
-        return s;
+
+        System.err.println("JNA Warning: Decoding with fallback " + System.getProperty("file.encoding"));
+        return new String(buf, 0, len);
     }
 
-    /** Obtain a Java String from the given native wchar_t array.  If there is
-     * no NUL terminator, the String will comprise the entire array.  
+    /**
+     * Obtain a Java String from the given native wchar_t array.  If there is
+     * no NUL terminator, the String will comprise the entire array.
+     *
+     * @param buf The buffer containing the characters
      */
     public static String toString(char[] buf) {
-        String s = new String(buf);
-        int term = s.indexOf(0);
-        if (term != -1) {
-            s = s.substring(0, term);
+        int len = buf.length;
+        for (int index = 0; index < len; index++) {
+            if (buf[index] == '\0') {
+                len = index;
+                break;
+            }
         }
-        return s;
+
+        if (len == 0) {
+            return "";
+        } else {
+            return new String(buf, 0, len);
+        }
+    }
+
+    /**
+     * Converts a &quot;list&quot; of strings each null terminated
+     * into a {@link List} of {@link String} values. The end of the
+     * list is signaled by an extra NULL value at the end or by the
+     * end of the buffer.
+     * @param buf The buffer containing the strings
+     * @return A {@link List} of all the strings in the buffer
+     * @see #toStringList(char[], int, int)
+     */
+    public static List<String> toStringList(char[] buf) {
+        return toStringList(buf, 0, buf.length);
+    }
+
+    /**
+     * Converts a &quot;list&quot; of strings each null terminated
+     * into a {@link List} of {@link String} values. The end of the
+     * list is signaled by an extra NULL value at the end or by the
+     * end of the data.
+     * @param buf The buffer containing the strings
+     * @param offset Offset to start parsing
+     * @param len The total characters to parse
+     * @return A {@link List} of all the strings in the buffer
+     */
+    public static List<String> toStringList(char[] buf, int offset, int len) {
+        List<String> list = new ArrayList<String>();
+        int lastPos = offset;
+        int maxPos = offset + len;
+        for (int curPos = offset; curPos < maxPos; curPos++) {
+            if (buf[curPos] != '\0') {
+                continue;
+            }
+
+            // check if found the extra null terminator
+            if (lastPos == curPos) {
+                return list;
+            }
+
+            String value = new String(buf, lastPos, curPos - lastPos);
+            list.add(value);
+            lastPos = curPos + 1;   // skip the '\0'
+        }
+
+        // This point is reached if there is no double null terminator
+        if (lastPos < maxPos) {
+            String value = new String(buf, lastPos, maxPos - lastPos);
+            list.add(value);
+        }
+
+        return list;
     }
 
     /** Map a library interface to the current process, providing
@@ -350,7 +440,7 @@ public final class Native implements Version {
      * <a href="NativeLibrary.html#library_search_paths">several locations</a>.
      * @param interfaceClass
      * @return an instance of the requested interface, mapped to the current
-     * process. 
+     * process.
      * @throws UnsatisfiedLinkError if the library cannot be found or
      * dependent libraries are missing.
      */
@@ -367,7 +457,7 @@ public final class Native implements Version {
      * @param interfaceClass
      * @param options Map of library options
      * @return an instance of the requested interface, mapped to the current
-     * process. 
+     * process.
      * @throws UnsatisfiedLinkError if the library cannot be found or
      * dependent libraries are missing.
      */
@@ -383,7 +473,7 @@ public final class Native implements Version {
      * @param name
      * @param interfaceClass
      * @return an instance of the requested interface, mapped to the indicated
-     * native library. 
+     * native library.
      * @throws UnsatisfiedLinkError if the library cannot be found or
      * dependent libraries are missing.
      */
@@ -402,7 +492,7 @@ public final class Native implements Version {
      * @param interfaceClass
      * @param options Map of library options
      * @return an instance of the requested interface, mapped to the indicated
-     * native library. 
+     * native library.
      * @throws UnsatisfiedLinkError if the library cannot be found or
      * dependent libraries are missing.
      */
@@ -549,7 +639,7 @@ public final class Native implements Version {
             return null;
         }
         catch (Exception e) {
-            throw new IllegalArgumentException(fieldName + " must be a public field of type " 
+            throw new IllegalArgumentException(fieldName + " must be a public field of type "
                                                + resultClass.getName() + " ("
                                                + e + "): " + mappingClass);
         }
@@ -562,9 +652,9 @@ public final class Native implements Version {
         return (TypeMapper)getLibraryOptions(cls).get(Library.OPTION_TYPE_MAPPER);
     }
 
-    /** Return the preferred Strring encoding for the given native interface. 
+    /** Return the preferred Strring encoding for the given native interface.
      * If there is no setting, defaults to the {@link
-     * #getDefaultStringEncoding()}. 
+     * #getDefaultStringEncoding()}.
      * See {@link com.sun.jna.Library#OPTION_STRING_ENCODING}.
      */
     public static String getStringEncoding(Class cls) {
@@ -596,7 +686,7 @@ public final class Native implements Version {
 
     /** Return a byte array corresponding to the given String, using the given
         encoding.  If the encoding is not found default to the platform native
-        encoding. 
+        encoding.
     */
     static byte[] getBytes(String s, String encoding) {
         if (encoding != null) {
@@ -665,20 +755,20 @@ public final class Native implements Version {
                 String dir = dirs.nextToken();
                 File file = new File(new File(dir), System.mapLibraryName(libName).replace(".dylib", ".jnilib"));
                 String path = file.getAbsolutePath();
-		if (DEBUG_JNA_LOAD) {
-		    System.out.println("Looking in " + path);
-		}
+                if (DEBUG_JNA_LOAD) {
+                    System.out.println("Looking in " + path);
+                }
                 if (file.exists()) {
                     try {
-			if (DEBUG_JNA_LOAD) {
-			    System.out.println("Trying " + path);
-			}
+                        if (DEBUG_JNA_LOAD) {
+                            System.out.println("Trying " + path);
+                        }
                         System.setProperty("jnidispatch.path", path);
                         System.load(path);
                         jnidispatchPath = path;
-			if (DEBUG_JNA_LOAD) {
-			    System.out.println("Found jnidispatch at " + path);
-			}
+                        if (DEBUG_JNA_LOAD) {
+                            System.out.println("Found jnidispatch at " + path);
+                        }
                         return;
                     } catch (UnsatisfiedLinkError ex) {
                         // Not a problem if already loaded in anoteher class loader
@@ -696,14 +786,14 @@ public final class Native implements Version {
                         ext = "dylib";
                     }
                     path = path.substring(0, path.lastIndexOf(orig)) + ext;
-		    if (DEBUG_JNA_LOAD) {
-			System.out.println("Looking in " + path);
-		    }
+                    if (DEBUG_JNA_LOAD) {
+                        System.out.println("Looking in " + path);
+                    }
                     if (new File(path).exists()) {
                         try {
-			    if (DEBUG_JNA_LOAD) {
-				System.out.println("Trying " + path);
-			    }
+                            if (DEBUG_JNA_LOAD) {
+                                System.out.println("Trying " + path);
+                            }
                             System.setProperty("jnidispatch.path", path);
                             System.load(path);
                             jnidispatchPath = path;
@@ -720,9 +810,9 @@ public final class Native implements Version {
         }
         if (!Boolean.getBoolean("jna.nosys")) {
             try {
-		if (DEBUG_JNA_LOAD) {
-		    System.out.println("Trying (via loadLibrary) " + libName);
-		}
+                if (DEBUG_JNA_LOAD) {
+                    System.out.println("Trying (via loadLibrary) " + libName);
+                }
                 System.loadLibrary(libName);
                 if (DEBUG_JNA_LOAD) {
                     System.out.println("Found jnidispatch on system path");
@@ -754,11 +844,11 @@ public final class Native implements Version {
                     throw new UnsatisfiedLinkError("Could not find JNA native support");
                 }
             }
-	    if (DEBUG_JNA_LOAD) {
-		System.out.println("Trying " + lib.getAbsolutePath());
-	    }
+            if (DEBUG_JNA_LOAD) {
+                System.out.println("Trying " + lib.getAbsolutePath());
+            }
             System.setProperty("jnidispatch.path", lib.getAbsolutePath());
-	    System.load(lib.getAbsolutePath());
+            System.load(lib.getAbsolutePath());
             jnidispatchPath = lib.getAbsolutePath();
             if (DEBUG_JNA_LOAD) {
                 System.out.println("Found jnidispatch at " + jnidispatchPath);
@@ -798,7 +888,7 @@ public final class Native implements Version {
     }
 
     /** Attempt to extract a native library from the resource path using the
-     * given class loader.  
+     * given class loader.
      * @param name Base name of native library to extract.  May also be an
      * absolute resource path (i.e. starts with "/"), in which case the
      * no transformations of the library name are performed.  If only the base
@@ -819,9 +909,9 @@ public final class Native implements Version {
                 loader = Native.class.getClassLoader();
             }
         }
-	if (DEBUG) {
-	    System.out.println("Looking in classpath from " + loader + " for " + name);
-	}
+        if (DEBUG) {
+            System.out.println("Looking in classpath from " + loader + " for " + name);
+        }
         String libname = name.startsWith("/") ? name : NativeLibrary.mapSharedLibraryName(name);
         String resourcePath = name.startsWith("/") ? name : Platform.RESOURCE_PREFIX + "/" + libname;
         if (resourcePath.startsWith("/")) {
@@ -851,9 +941,9 @@ public final class Native implements Version {
             catch(URISyntaxException e) {
                 lib = new File(url.getPath());
             }
-	    if (DEBUG) {
-		System.out.println("Looking in " + lib.getAbsolutePath());
-	    }
+            if (DEBUG) {
+                System.out.println("Looking in " + lib.getAbsolutePath());
+            }
             if (!lib.exists()) {
                 throw new IOException("File URL " + url + " could not be properly decoded");
             }
@@ -903,20 +993,23 @@ public final class Native implements Version {
     private static native String getNativeVersion();
     private static native String getAPIChecksum();
 
-    /** Retrieve the last error set by the OS.  This corresponds to
+    /** Retrieve last error set by the OS.  This corresponds to
      * <code>GetLastError()</code> on Windows, and <code>errno</code> on
      * most other platforms.  The value is preserved per-thread, but whether
-     * the original value is per-thread depends on the underlying OS.  The
-     * result is undefined if {@link #getPreserveLastError} is
-     * <code>false</code>.<p>
-     * The preferred method of obtaining the last error result is
+     * the original value is per-thread depends on the underlying OS.  
+     * <p>
+     * An alternative method of obtaining the last error result is
      * to declare your mapped method to throw {@link LastErrorException}
-     * instead.
+     * instead.  If a method's signature includes a throw of {@link
+     * LastErrorException}, the last error will be set to zero before the
+     * native call and a {@link LastErrorException} will be raised if the last
+     * error value is non-zero after the call, regardless of the actual
+     * returned value from the native function.</p>
      */
     public static native int getLastError();
 
     /** Set the OS last error code.  The value will be saved on a per-thread
-     * basis. 
+     * basis.
      */
     public static native void setLastError(int code);
 
@@ -941,6 +1034,7 @@ public final class Native implements Version {
         }
         final Library.Handler handler = (Library.Handler)ih;
         InvocationHandler newHandler = new InvocationHandler() {
+            @Override
             public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                 synchronized(handler.getNativeLibrary()) {
                     return handler.invoke(library, method, args);
@@ -974,6 +1068,7 @@ public final class Native implements Version {
 
             final ClassLoader cl = Native.class.getClassLoader();
             Method m = (Method)AccessController.doPrivileged(new PrivilegedAction() {
+                @Override
                 public Object run() {
                     try {
                         Method m = ClassLoader.class.getDeclaredMethod("findLibrary", new Class[] { String.class });
@@ -1042,6 +1137,7 @@ public final class Native implements Version {
     static void removeTemporaryFiles() throws IOException {
         File dir = getTempDir();
         FilenameFilter filter = new FilenameFilter() {
+            @Override
             public boolean accept(File dir, String name) {
                 return name.endsWith(".x") && name.startsWith(JNA_TMPLIB_PREFIX);
             }
@@ -1190,6 +1286,7 @@ public final class Native implements Version {
     */
     static Class getCallingClass() {
         Class[] context = new SecurityManager() {
+            @Override
             public Class[] getClassContext() {
                 return super.getClassContext();
             }
@@ -1213,19 +1310,18 @@ public final class Native implements Version {
     }
 
 
-    private static Map registeredClasses = new HashMap();
-    private static Map registeredLibraries = new HashMap();
-    private static Object unloader = new Object() {
-        protected void finalize() {
-            synchronized(registeredClasses) {
-                for (Iterator i=registeredClasses.entrySet().iterator();i.hasNext();) {
-                    Map.Entry e = (Map.Entry)i.next();
-                    unregister((Class)e.getKey(), (long[])e.getValue());
-                    i.remove();
-                }
+    private static Map registeredClasses = new WeakHashMap();
+    private static Map registeredLibraries = new WeakHashMap();
+
+    private static void unregisterAll() {
+        synchronized(registeredClasses) {
+            for (Iterator i=registeredClasses.entrySet().iterator();i.hasNext();) {
+                Map.Entry e = (Map.Entry)i.next();
+                unregister((Class)e.getKey(), (long[])e.getValue());
+                i.remove();
             }
         }
-    };
+    }
 
     /** Remove all native mappings for the calling class.
         Should only be called if the class is no longer referenced and about
@@ -1249,6 +1345,13 @@ public final class Native implements Version {
         }
     }
 
+    /** @return whether the given class's native components are registered. */
+    public static boolean registered(Class cls) {
+        synchronized(registeredClasses) {
+            return registeredClasses.containsKey(cls);
+        }
+    }
+
     /** Unregister the native methods for the given class. */
     private static native void unregister(Class cls, long[] handles);
 
@@ -1310,10 +1413,14 @@ public final class Native implements Version {
     private static final int CVT_CALLBACK = 15;
     private static final int CVT_FLOAT = 16;
     private static final int CVT_NATIVE_MAPPED = 17;
-    private static final int CVT_WSTRING = 18;
-    private static final int CVT_INTEGER_TYPE = 19;
-    private static final int CVT_POINTER_TYPE = 20;
-    private static final int CVT_TYPE_MAPPER = 21;
+    private static final int CVT_NATIVE_MAPPED_STRING = 18;
+    private static final int CVT_NATIVE_MAPPED_WSTRING = 19;
+    private static final int CVT_WSTRING = 20;
+    private static final int CVT_INTEGER_TYPE = 21;
+    private static final int CVT_POINTER_TYPE = 22;
+    private static final int CVT_TYPE_MAPPER = 23;
+    private static final int CVT_TYPE_MAPPER_STRING = 24;
+    private static final int CVT_TYPE_MAPPER_WSTRING = 25;
 
     private static int getConversion(Class type, TypeMapper mapper) {
         if (type == Boolean.class) type = boolean.class;
@@ -1326,10 +1433,29 @@ public final class Native implements Version {
         else if (type == Double.class) type = double.class;
         else if (type == Void.class) type = void.class;
 
-        if (mapper != null
-            && (mapper.getFromNativeConverter(type) != null
-                || mapper.getToNativeConverter(type) != null)) {
-            return CVT_TYPE_MAPPER;
+        if (mapper != null) {
+            FromNativeConverter fromNative = mapper.getFromNativeConverter(type);
+            ToNativeConverter toNative = mapper.getToNativeConverter(type);
+            if (fromNative != null) {
+                Class nativeType = fromNative.nativeType();
+                if (nativeType == String.class) {
+                    return CVT_TYPE_MAPPER_STRING;
+                }
+                if (nativeType == WString.class) {
+                    return CVT_TYPE_MAPPER_WSTRING;
+                }
+                return CVT_TYPE_MAPPER;
+            }
+            if (toNative != null) {
+                Class nativeType = toNative.nativeType();
+                if (nativeType == String.class) {
+                    return CVT_TYPE_MAPPER_STRING;
+                }
+                if (nativeType == WString.class) {
+                    return CVT_TYPE_MAPPER_WSTRING;
+                }
+                return CVT_TYPE_MAPPER;
+            }
         }
 
         if (Pointer.class.isAssignableFrom(type)) {
@@ -1376,6 +1502,13 @@ public final class Native implements Version {
             return CVT_POINTER_TYPE;
         }
         if (NativeMapped.class.isAssignableFrom(type)) {
+            Class nativeType = NativeMappedConverter.getInstance(type).nativeType();
+            if (nativeType == String.class) {
+                return CVT_NATIVE_MAPPED_STRING;
+            }
+            if (nativeType == WString.class) {
+                return CVT_NATIVE_MAPPED_WSTRING;
+            }
             return CVT_NATIVE_MAPPED;
         }
         return CVT_UNSUPPORTED;
@@ -1388,7 +1521,7 @@ public final class Native implements Version {
      * library load path or <code>jna.library.path</code>.
      * @param cls Class with native methods to register
      * @param libName name of or path to native library to which functions
-     * should be bound 
+     * should be bound
      */
     public static void register(Class cls, String libName) {
         Map options = new HashMap();
@@ -1410,6 +1543,7 @@ public final class Native implements Version {
         List mlist = new ArrayList();
         TypeMapper mapper = (TypeMapper)
             lib.getOptions().get(Library.OPTION_TYPE_MAPPER);
+        cacheOptions(cls, lib.getOptions(), null);
 
         for (int i=0;i < methods.length;i++) {
             if ((methods[i].getModifiers() & Modifier.NATIVE) != 0) {
@@ -1434,11 +1568,18 @@ public final class Native implements Version {
             case CVT_UNSUPPORTED:
                 throw new IllegalArgumentException(rclass + " is not a supported return type (in method " + method.getName() + " in " + cls + ")");
             case CVT_TYPE_MAPPER:
+            case CVT_TYPE_MAPPER_STRING:
+            case CVT_TYPE_MAPPER_WSTRING:
                 fromNative = mapper.getFromNativeConverter(rclass);
-                closure_rtype = FFIType.get(rclass).peer;
+                // FFIType.get() always looks up the native type for any given
+                // class, so if we actually have conversion into a Java
+                // object, make sure we use the proper type information
+                closure_rtype = FFIType.get(rclass.isPrimitive() ? rclass : Pointer.class).peer;
                 rtype = FFIType.get(fromNative.nativeType()).peer;
                 break;
             case CVT_NATIVE_MAPPED:
+            case CVT_NATIVE_MAPPED_STRING:
+            case CVT_NATIVE_MAPPED_WSTRING:
             case CVT_INTEGER_TYPE:
             case CVT_POINTER_TYPE:
                 closure_rtype = FFIType.get(Pointer.class).peer;
@@ -1463,10 +1604,14 @@ public final class Native implements Version {
                     throw new IllegalArgumentException(type + " is not a supported argument type (in method " + method.getName() + " in " + cls + ")");
                 }
                 if (cvt[t] == CVT_NATIVE_MAPPED
+                    || cvt[t] == CVT_NATIVE_MAPPED_STRING
+                    || cvt[t] == CVT_NATIVE_MAPPED_WSTRING
                     || cvt[t] == CVT_INTEGER_TYPE) {
                     type = NativeMappedConverter.getInstance(type).nativeType();
                 }
-                else if (cvt[t] == CVT_TYPE_MAPPER) {
+                else if (cvt[t] == CVT_TYPE_MAPPER
+                         || cvt[t] == CVT_TYPE_MAPPER_STRING
+                         || cvt[t] == CVT_TYPE_MAPPER_WSTRING) {
                     toNative[t] = mapper.getToNativeConverter(type);
                 }
                 // Determine the type that will be passed to the native
@@ -1477,14 +1622,15 @@ public final class Native implements Version {
                 case CVT_INTEGER_TYPE:
                 case CVT_POINTER_TYPE:
                 case CVT_NATIVE_MAPPED:
+                case CVT_NATIVE_MAPPED_STRING:
+                case CVT_NATIVE_MAPPED_WSTRING:
                     atypes[t] = FFIType.get(type).peer;
                     closure_atypes[t] = FFIType.get(Pointer.class).peer;
                     break;
                 case CVT_TYPE_MAPPER:
-                    if (type.isPrimitive())
-                        closure_atypes[t] = FFIType.get(type).peer;
-                    else
-                        closure_atypes[t] = FFIType.get(Pointer.class).peer;
+                case CVT_TYPE_MAPPER_STRING:
+                case CVT_TYPE_MAPPER_WSTRING:
+                    closure_atypes[t] = FFIType.get(type.isPrimitive() ? type : Pointer.class).peer;
                     atypes[t] = FFIType.get(toNative[t].nativeType()).peer;
                     break;
                 case CVT_DEFAULT:
@@ -1526,7 +1672,6 @@ public final class Native implements Version {
             registeredClasses.put(cls, handles);
             registeredLibraries.put(cls, lib);
         }
-        cacheOptions(cls, lib.getOptions(), null);
     }
 
     /** Take note of options used for a given library mapping, to facilitate
@@ -1655,10 +1800,10 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      *
-     * @return	The value returned by the target native function
+     * @return  The value returned by the target native function
      */
     static  native int invokeInt(long fp, int callFlags, Object[] args);
 
@@ -1666,10 +1811,10 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      *
-     * @return	The value returned by the target native function
+     * @return  The value returned by the target native function
      */
     static native long invokeLong(long fp, int callFlags, Object[] args);
 
@@ -1677,8 +1822,8 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      */
     static native void invokeVoid(long fp, int callFlags, Object[] args);
 
@@ -1686,10 +1831,10 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      *
-     * @return	The value returned by the target native function
+     * @return  The value returned by the target native function
      */
     static native float invokeFloat(long fp, int callFlags, Object[] args);
 
@@ -1697,10 +1842,10 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      *
-     * @return	The value returned by the target native function
+     * @return  The value returned by the target native function
      */
     static native double invokeDouble(long fp, int callFlags, Object[] args);
 
@@ -1708,10 +1853,10 @@ public final class Native implements Version {
      * Call the native function being represented by this object
      * @param fp function pointer
      * @param   callFlags calling convention to be used
-     * @param	args
-     *			Arguments to pass to the native function
+     * @param   args
+     *                  Arguments to pass to the native function
      *
-     * @return	The native pointer returned by the target native function
+     * @return  The native pointer returned by the target native function
      */
     static native long invokePointer(long fp, int callFlags, Object[] args);
 
@@ -1885,11 +2030,11 @@ public final class Native implements Version {
      */
     public static native ByteBuffer getDirectByteBuffer(long addr, long length);
 
-    /** Indicate whether the JVM should detach the current native thread when
+    /** <p>Indicate whether the JVM should detach the current native thread when
         the current Java code finishes execution.  Generally this is used to
         avoid detaching native threads when it is known that a given thread
         will be relatively long-lived and call back to Java code frequently.
-        <p/>
+        </p>
         This call is lightweight; it only results in an additional JNI
         crossing if the desired state changes from its last setting.
 
@@ -1924,8 +2069,9 @@ public final class Native implements Version {
 
     private static Map nativeThreads = Collections.synchronizedMap(new WeakHashMap());
 
-    private static ThreadLocal nativeThreadTerminationFlag = 
+    private static ThreadLocal nativeThreadTerminationFlag =
         new ThreadLocal() {
+            @Override
             protected Object initialValue() {
                 Memory m = new Memory(4);
                 m.clear();
