@@ -1,4 +1,4 @@
-/* Copyright (c) 2009 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2009-2015 Timothy Wall, All Rights Reserved
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -20,6 +20,9 @@ import java.io.File;
 import java.nio.Buffer;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.util.Map;
+import java.util.HashMap;
+import java.lang.reflect.Method; 
 
 import com.sun.jna.DirectTest.TestInterface;
 import com.sun.jna.DirectTest.TestLibrary;
@@ -29,7 +32,7 @@ public class PerformanceTest extends TestCase implements Paths {
 
     public void testEmpty() { }
 
-    private static class JNI {
+    private static class JNILibrary {
         static {
             String path = TESTPATH + NativeLibrary.mapSharedLibraryName("testlib");;
             if (!new File(path).isAbsolute()) {
@@ -39,6 +42,7 @@ public class PerformanceTest extends TestCase implements Paths {
         }
         
         private static native double cos(double x);
+        private static native int getpid();
     }
 
     public static void main(java.lang.String[] argList) {
@@ -83,7 +87,22 @@ public class PerformanceTest extends TestCase implements Paths {
         }
     }
 
+    static class PIDLibrary {
+        public static native int getpid();
+        static {
+            Native.register(Platform.C_LIBRARY_NAME);
+        }
+    }
+
+    static class W32PIDLibrary {
+        public static native int _getpid();
+        static {
+            Native.register(Platform.C_LIBRARY_NAME);
+        }
+    }
+
     static interface CInterface extends Library {
+        int getpid();
         Pointer memset(Pointer p, int v, int len);
         int strlen(String s);
     }
@@ -168,7 +187,7 @@ public class PerformanceTest extends TestCase implements Paths {
 
         start = System.currentTimeMillis();
         for (int i=0;i < COUNT;i++) {
-            dresult = JNI.cos(0d);
+            dresult = JNILibrary.cos(0d);
         }
         delta = System.currentTimeMillis() - start;
         System.out.println("cos (JNI): " + delta + "ms");
@@ -180,12 +199,56 @@ public class PerformanceTest extends TestCase implements Paths {
         delta = System.currentTimeMillis() - start;
         System.out.println("cos (pure java): " + delta + "ms");
 
-        ///////////////////////////////////////////
-        // memset
         Pointer presult;
         String cname = Platform.C_LIBRARY_NAME;
+        Map options = new HashMap();
+        if (Platform.isWindows()) {
+            options.put(Library.OPTION_FUNCTION_MAPPER, new FunctionMapper() {
+                public String getFunctionName(NativeLibrary library, Method method) {
+                    String name = method.getName();
+                    if ("getpid".equals(name)) { 
+                        name = "_getpid";
+                    }
+                    return name;
+                }
+            });
+        }
         CInterface clib = (CInterface)
-            Native.loadLibrary(cname, CInterface.class);
+            Native.loadLibrary(cname, CInterface.class, options);
+
+        ///////////////////////////////////////////
+        // getpid
+        int pid;
+        start = System.currentTimeMillis();
+        for (int i=0;i < COUNT;i++) {
+            pid = clib.getpid();
+        }
+        delta = System.currentTimeMillis() - start;
+        System.out.println("getpid (JNA interface): " + delta + "ms");
+
+        start = System.currentTimeMillis();
+        if (Platform.isWindows()) {
+            for (int i=0;i < COUNT;i++) {
+                pid = W32PIDLibrary._getpid();
+            }
+        }
+        else {
+            for (int i=0;i < COUNT;i++) {
+                pid = PIDLibrary.getpid();
+            }
+        }
+        delta = System.currentTimeMillis() - start;
+        System.out.println("getpid (JNA direct): " + delta + "ms");
+
+        start = System.currentTimeMillis();
+        for (int i=0;i < COUNT;i++) {
+            pid = JNILibrary.getpid();
+        }
+        delta = System.currentTimeMillis() - start;
+        System.out.println("getpid (JNI): " + delta + "ms");
+
+        ///////////////////////////////////////////
+        // memset
         start = System.currentTimeMillis();
         for (int i=0;i < COUNT;i++) {
             presult = clib.memset(null, 0, 0);
